"""

Test config construction

Simple tests that ensure that the Config construction isn't broken

"""

import unittest
import logging

logger = logging.getLogger("basic_construct")
logger.setLevel(level=logging.INFO)

import configerus

class BasicConstruct(unittest.TestCase):

    def test_construct_1_naked(self):
        """ just make sure we can create a Configs object, without any bootstraps """
        config = configerus.new_config(bootstraps=[])

    def test_construct_2_single_bootstrap(self):
        """ make sure we can create a Configs object, without a single bootstrap """
        config = configerus.new_config(bootstraps=['dict'])

    def test_construct_3_bootsraps(self):
        """ Make sure we can create a Configs object, with default bootstraps """
        config = configerus.new_config()

if __name__ == '__main__':
    unittest.main()



from jsonschema import validate

from configerus.config import Config

PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL = 'jsonschema'

class JsonSchemaValidatorPlugin():
    """   """

    def __init__(self, config:Config, instance_id:str):
        """  """
        self.config = config
        self.instance_id = config

    def validate(self, validate_target:str, data):
        """ Validate a structure using jsonschema pulled from a config key """

        (method, validate_key) = validate_target.split(':')

        if not method == PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL:
            # returns of any value signal validation as validate only catches exceptions
            return

        try:
            schema_config = self.config.load(PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL)
            schema = schema_config.get(validate_key, exception_if_missing=True)

        except Exception as e:
            raise NotImplemented("Could not access jsonschema validation schema from config target '{}:{}'".format(PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL, validate_key))

        validate(data, schema=schema)



"""

Test config templating

Here we run some simple config template checking to make sure that template
substitution of whole and partial values works.

"""
import logging
import unittest

import configerus
from configerus.contrib.dict import PLUGIN_ID_CONFIGSOURCE_DICT
from configerus.contrib.files import PLUGIN_ID_CONFIGSOURCE_PATH

from configerus.test import make_test_config, test_config_cleanup

logger = logging.getLogger("test_config_behaviour")
logger.setLevel(logging.INFO)

config_sources = [
    {
        "name": "source",
        "priority": 30,
        "type": PLUGIN_ID_CONFIGSOURCE_DICT,
        "data": {
            "config": {
                "1": "first 1",
                "2": {
                    "1": "first 2.1"
                },
                "3": {
                    "1": {
                        "1": "first 3.1.1",
                        "2": "first 3.1.2"
                    },
                    "2": {
                        "1": "first 3.2.1",
                        "2": "first 3.2.2"
                    }
                }
            },
            "variables": {
                "one": "first one",
                "two": "first two",
                "three": 3,
                "four": {
                    "1": "first four.1"
                }
            }
        }
    },
    {
        "name": "simple",
        "priority": 80,
        "type": PLUGIN_ID_CONFIGSOURCE_DICT,
        "data": {
            "config": {
                "10": "{2}",
                "11": "__{2}__"
            }
        }
    },
    {
        "name": "defaults",
        "priority": 80,
        "type": PLUGIN_ID_CONFIGSOURCE_DICT,
        "data": {
            "config": {
                "20": "{does.not.exist?default}",
                "21": "{does.not.exist}",
                "22": "{variables:does.not.exist?megadefault}",
                "23": "{variables:three?default}",
            }
        }
    },
    {
        "name": "special",
        "priority": 80,
        "type": PLUGIN_ID_CONFIGSOURCE_DICT,
        "data": {
            "config": {
                "30": "{path:replace}",
            }
        }
    },
    {
        "name": "deep",
        "priority": 80,
        "type": PLUGIN_ID_CONFIGSOURCE_DICT,
        "data": {
            "config": {
                "40": "deep 40",
                "41": "{40}",
                "42": {
                    "1": "{40}",
                    "2": {
                        "1": "{40}"
                    }
                }
            }
        }
    },
    {
        "name": "replace",
        "priority": 80,
        "type": PLUGIN_ID_CONFIGSOURCE_PATH,
        "data": {
            "replace.json": {
                "id": "test"
            }
        }
    }
]
""" Contents of test config files used as the source for a config object """

""" TESTS """

class ConfigTemplating(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ make a Config test object from some inline data

        First this dumps the data into json files, and then points the config object
        to the various paths for loading. The data can then be used to test funct.
        The data contains only values that are usefull for confirming that config
        behaviour is as expected, and is not meant to be useful.

        We also preload some of the config
        """

        logger.debug("Building empty config object")
        config = configerus.new_config()
        make_test_config(config, config_sources)
        cls.config = config

        cls.loaded_config = config.load("config")
        cls.loaded_variables =config.load("variables")

    @classmethod
    def tearDownClass(cls):
        test_config_cleanup(cls.config)

    def test_format_sane(self):
        """ simple template replacing using format """

        assert self.config.format('{1}', 'config') == "first 1"
        assert self.loaded_config.format('{1}') == "first 1"
        """ Simple formatting at the config and loadedconfig level """


    def test_format_dot_notation(self):
        """ test that format catches dot notation """

        assert self.config.format('{2.1}', 'config') == "first 2.1"
        assert self.config.format('{3.2.1}', 'config') == "first 3.2.1"
        """ Simple dot notation descending format """

    def test_missing_raises_exceptions(self):
        """ missing values should raise exceptions """

        try:
            assert self.config.format("{does.not.exist}", 'config') == "why-no-exception?"
        except KeyError:
            return

        assert False, "Expected exception"

    def test_format_embedding_and_casting(self):
        """ simple template replacing using format """

        assert self.config.format('12345{1}67890', 'config') == "12345first 167890"
        assert self.config.format('1 23 45 {1}6 78 90', 'config') == "1 23 45 first 16 78 90"
        """ format replacements in the middle of a string """

        assert self.config.format('{2}', 'config') == {"1": "first 2.1"}
        assert self.config.format(' {2} ', 'config') == " {'1': 'first 2.1'} "
        """ format should string cast only not full matches """

    def test_format_with_sources(self):
        """ test that sources work with format """

        variables_two = self.config.format("{two}", 'variables')
        assert variables_two, "Except value for comparison"
        assert self.config.format("{variables:two}", 'config') == variables_two
        assert self.config.format("{variables:two}", 'variables') == variables_two
        """ specifying a source should work """

    def test_format_with_defaults(self):
        """ test the defaulting with formatting """

        assert self.config.format("{does.not.exist?default1}", 'config') == 'default1'
        assert self.config.format("{variables:does.not.exist?default2}", 'config') == 'default2'
        """ defaults should apply """

        variables_two = self.config.format("{two}", 'variables')
        assert variables_two, "Except value for comparison"
        assert self.config.format("{variables:two?default}", 'config') == variables_two
        """ defaults should not overwrite values that can be found """

    def test_get_formatted(self):
        """ test that loadedconfig gets format """

        config_one =self.loaded_config.get("1")
        assert config_one, "Expected value for comparison"
        assert self.config.format('{1}', 'config') == config_one
        """ get should mimic format behaviour, which we tested above """

        variables_three =self.loaded_variables.get("three")
        assert config_one, "Expected value for comparison"
        assert self.loaded_config.format("{variables:three}") == variables_three
        """ test templating for different sources """

        assert self.loaded_config.get("20") == 'default'
        assert self.loaded_config.get("22") == 'megadefault'
        assert self.loaded_config.get("23") == variables_three

        try:
            assert self.loaded_config.get("21")
        except:
            return
        assert False, "expected exception"

    def test_deep_formatting(self):
        """ test that formatting is applied deeply """

        config_fourty =self.loaded_config.get("40")
        assert config_fourty, "Expected value for comparison"

        assert self.loaded_config.get("41") == config_fourty
        assert self.loaded_config.get("42")["1"] == config_fourty
        assert self.loaded_config.get("42")["2"]["1"] == config_fourty




import functools
import logging
import copy
from enum import Enum, unique

logger = logging.getLogger('configerus.plugin')
# logger.setLevel(level=logging.DEBUG)

@unique
class Type(Enum):
    """ Enumerator to match plugin types to plugin labels """
    CONFIGSOURCE = "configerus.plugin.configsource"
    """ A Config source handler """
    FORMATTER    = "configerus.plugin.formatter"
    """ A string formatter plugin """
    VALIDATOR    = "configerus.plugin.validator"
    """ A data result validator plugin """

class Factory():
    """ Python decorator class for configuerus Plugin factories

    This class should be used to decorate any function which is meant to be a
    factory for plugins.

    If you are writing a plugin factory, decorate it with this class, provide
    the factory type and id values, and then the factory will be avaialble to
    other code.

    If you are trying to get an instance of a plugin, then create an instance of
    this class and use the create() method
    """

    registry = {}
    """ A static list of all of the registered factory functions """

    def __init__(self, type: Type, plugin_id: str):
        """ register the decoration """
        self.plugin_id = plugin_id
        self.type = type

        if not self.type.value in self.registry:
            self.registry[self.type.value] = {}

    def __call__(self, func):
        """ call the decorated function

        Returns:

        wrapped function(config: Config)
        """
        def wrapper(config, instance_id: str):
            logger.debug("plugin factory exec: %s:%s", self.type.value, self.plugin_id)
            return func(config=config, instance_id=instance_id)

        logger.debug("Plugin factory registered `%s:%s`", self.type.value, self.plugin_id)
        self.registry[self.type.value][self.plugin_id] = wrapper
        return wrapper

    def create(self, config, instance_id: str):
        """ Get an instance of a plugin as created by the decorated """
        try:
            factory = self.registry[self.type.value][self.plugin_id]
        except KeyError:
            raise NotImplementedError("Configerus Plugin instance '{}:{}' has not been registered.".format(self.type.value, self.plugin_id))
        # except Exception as e:
        #     raise Exception("Could not create Plugin instance '{}:{}' as the plugin factory produced an exception".format(self.type.value, self.plugin_id)) from e

        return factory(config=config, instance_id=instance_id)


CONFIGERUS_FORMATTER_TYPE = Type.FORMATTER
""" Short cut to the config source plugin type enum """
class FormatFactory(Factory):
    """ Decoration class for registering a config source plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """
    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_FORMATTER_TYPE, plugin_id)


CONFIGERUS_SOURCE_TYPE = Type.CONFIGSOURCE
""" Short cut to the config source plugin type enum """
class SourceFactory(Factory):
    """ Decoration class for registering a config source plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """
    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_SOURCE_TYPE, plugin_id)


CONFIGERUS_VALIDATOR_TYPE = Type.VALIDATOR
""" Short cut to the validate plugin type enum """
class ValidatorFactory(Factory):
    """ Decoration class for registering a validate plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """
    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_VALIDATOR_TYPE, plugin_id)


class PluginInstances:
    """ List of plugins wrapped in the PluginInstance struct so that it can be
        prioritized and managed.
    """

    def __init__(self, config):
        self.instances = []
        """ keep a list of all of the plugins as PluginInstance wrappers

            This mixes plugin types together but but it simplifies management
            and storage of plugins """

        self.config = config
        """ A config object is needed for the plugin factories, and it doesn't
            make sense to use a different config for different plugins """

    def copy(self, new_config):
        """ Make a copy of this plugin list.

        This copies the instance list and by copying every plugin, overriding
        its config and adding it to the new list

        @TODO this is not really very good and should be fixed

        """
        instances_copy = PluginInstances(new_config)
        """A new instance list which we will return after copying over plugins """

        for instance in self.instances:
            plugin_copy = copy.deepcopy(instance.plugin)
            plugin_copy.config = new_config
            instances_copy.instances.append(PluginInstance(instance.type, instance.instance_id, instance.priority, plugin_copy))
        return instances_copy

    def add_plugin(self, type:Type, plugin_id:str, instance_id:str, priority:int):
        """ Create a configerus plugin and add it to the config object

        Parameters:
        -----------

        plugin_id (str) : id of the plugin as registered using the plugin
            factory decorater. This has to match a plugin's registration with

        instance_id (str) : Optionally give a plugin instance a name, which it
            might use for internal functionality.
            The "path" source plugin allows string template substitution of the
            "__paths__:instance_id" for the path.

        priority (int) : plugin priority. Use this to set this plugins as
            higher or lower priority than others.

        Returns:
        --------

        Returns the plugin so that you can do any actions to the plugin that it
        supports, and the code here doesn't need to get fancy with function
        arguments

        """
        if not plugin_id:
            raise KeyError("Could not create a plugin as an invalid plugin_id was given: '{}'".format(plugin_id))

        if not type:
            raise KeyError("Could not create a plugin as an invalid type was given: '{}'".format(plugin_id))

        if not instance_id:
            # generate some kind of unique instance_id key
            base_instance_id = "{}_{}".format(type.value, plugin_id)
            index = 1
            instance_id = "{}_{}".format(base_instance_id, index)
            while self.has_plugin(instance_id):
                index = 1
                instance_id = "{}_{}".format(base_instance_id, index)

        if self.has_plugin(instance_id, plugin_id):
            self.warn("Replacing '{}.{}' with new plugin instance".format(plugin_id, instance_id))

        try:
            fac = Factory(type, plugin_id)
            plugin = fac.create(self.config, instance_id)
            instance = PluginInstance(type, instance_id, priority, plugin)

        except NotImplementedError as e:
            raise NotImplementedError("Could not create configerus plugin '{}:{}' as that plugin_id could not be found.".format(type.value, plugin_id)) from e

        self.instances.append(instance)
        return plugin

    def get_plugin(self, instance_id, type:Type=None, exception_if_missing: bool=True):
        """ Retrieve a plugin from its instance_id, optionally of a specific type """
        for plugin_instance in self.instances:
            if plugin_instance.instance_id==instance_id and (type==None or type==plugin_instance.type):
                return plugin_instance.plugin

        if exception_if_missing:
            raise KeyError("Could not find plugin {}".format(instance_id if type is None else "{}:{}".format(type.value, instance_id)))
        return False

    def has_plugin(self, instance_id, type:Type=None):
        """ Discover if a plugin had been added with an instance_id, optionally
            of a specific type """
        return bool(self.get_plugin(instance_id, type, exception_if_missing=False))

    def get_ordered_plugins(self, type:Type=None):
        """ order plugin, optionally just one type, by their top down priority

        we process the plugins in three steps:
        1. Optionally filter for type
        2. sort based on priority
        3. collect just the instance plugin objects

        """
        typed_instances = [instance for instance in self.instances if (type is None or instance.type==type)]
        sorted_instances = sorted(typed_instances, key=lambda instance: instance.priority)
        sorted_instances.reverse()
        return [instance.plugin for instance in sorted_instances]

class PluginInstance:
    """ a struct for a plugin instance that also keeps metadata about the instance """

    def __init__(self, type:Type, instance_id:str, priority:int, plugin):
        self.type = type
        self.instance_id = instance_id
        self.priority = priority
        self.plugin = plugin



"""

Test config's copy()

Here we test that the Config object is able to make copies of itself, and that
changing those copies doesn't affect originals.

"""

import unittest
import logging

logging.basicConfig()
logger = logging.getLogger('config_copy')
logger.setLevel(level=logging.INFO)

import configerus
from configerus.contrib.dict import PLUGIN_ID_CONFIGSOURCE_DICT

""" TESTS """

class ConfigBehaviour(unittest.TestCase):

    def test_copy_safety(self):
        """ Test that config copy allows overloads and doesn't modify the source """

        logger.debug("Building empty config object")
        config = configerus.new_config()

        config.add_source(PLUGIN_ID_CONFIGSOURCE_DICT, 'orig', 80).set_data({
            'copy': {
                'one': 'orig 1'
            }
        })

        config_copy_orig = config.load('copy')

        copy1 = config.copy()
        copy1.add_source(PLUGIN_ID_CONFIGSOURCE_DICT, 'copy1', 80).set_data({
            'copy': {
                'one': 'copy1 1',
                'two': 'copy1 2'
            }
        })
        copy2 = config.copy()
        copy2.add_source(PLUGIN_ID_CONFIGSOURCE_DICT, 'copy2', 81).set_data({
            'copy': {
                'one': 'copy2 1',
                'two': 'copy2 2'
            }
        })

        config_copy_late = config.load('copy')
        config1_copy = copy1.load('copy')
        config2_copy = copy2.load('copy')

        logger.debug('LOADED: AFTER: orig: %s',config_copy_orig.data )
        logger.debug('LOADED: AFTER: 1   : %s',config1_copy.data )
        logger.debug('LOADED: AFTER: 2   : %s',config2_copy.data )
        logger.debug('LOADED: AFTER: late: %s',config_copy_late.data )

        # check original values
        assert config_copy_orig.get('one') == 'orig 1'
        assert config_copy_orig.get('two') == None
        # check that copied config didn't modify original
        assert config_copy_orig.get('one') == config_copy_late.get('one')
        assert config_copy_orig.get('two') == config_copy_late.get('two')

        assert config1_copy.get('one') == 'copy1 1'
        assert config1_copy.get('two') == 'copy1 2'
